"use strict";
var test = require("tape");
var runtime_1 = require("../src/runtime/runtime");
var parser = require("../src/runtime/parser");
var builder = require("../src/runtime/builder");
function dedent(str) {
    var lines = [];
    var indent;
    for (var _i = 0, _a = str.split("\n"); _i < _a.length; _i++) {
        var line = _a[_i];
        var match = line.match(/^[ \t]+/);
        if (match) {
            if (!indent) {
                indent = match[0].length;
            }
            line = line.substr(indent);
        }
        lines.push(line);
    }
    return lines.join("\n");
}
exports.dedent = dedent;
function eavsToComparables(eavs, entities, index) {
    if (index === void 0) { index = {}; }
    var results = [];
    for (var _i = 0, eavs_1 = eavs; _i < eavs_1.length; _i++) {
        var eav = eavs_1[_i];
        var e = eav[0], a = eav[1], v = eav[2];
        var cur = index[e];
        if (!index[e]) {
            cur = index[e] = { list: [], links: [], e: e };
            results.push(cur);
        }
        if (entities[v]) {
            cur.links.push([a, v]);
        }
        else {
            var avKey = a + ", " + v;
            cur.list.push(avKey);
        }
    }
    return results;
}
exports.eavsToComparables = eavsToComparables;
// Tests if the elements of `as` are in `bs` by comparing keys in a list. In
// the case of floating points, we want the comparison to be looser, to account
// for floating point rounding errors, which can vary by platform.
function isSetEqual(as, bs, assert) {
    if (as.length !== bs.length)
        return false;
    for (var _i = 0, as_1 = as; _i < as_1.length; _i++) {
        var a = as_1[_i];
        var avA = a.split(", ");
        var attributeA = avA[0];
        var valueA = avA[1];
        // If the attribute is called "floatTest" we know to compare
        // the values to within some epsilon. Otherwise, we can
        // do a straight lookup of the key
        var epsilon = 0.00000000001;
        if (attributeA === "floatTest") {
            for (var _a = 0, bs_1 = bs; _a < bs_1.length; _a++) {
                var b = bs_1[_a];
                var avB = b.split(", ");
                var attributeB = avB[0];
                var valueB = avB[1];
                if (attributeB === "floatTest" && Math.abs(valueB - valueA) < epsilon) {
                    return true;
                }
            }
            return false;
        }
        else {
            if (bs.indexOf(a) === -1)
                return false;
        }
    }
    return true;
}
exports.isSetEqual = isSetEqual;
function collectEntities(eavs, index) {
    if (index === void 0) { index = {}; }
    for (var _i = 0, eavs_2 = eavs; _i < eavs_2.length; _i++) {
        var e = eavs_2[_i][0];
        index[e] = true;
    }
    return index;
}
var Resolution;
(function (Resolution) {
    Resolution[Resolution["unknown"] = 0] = "unknown";
    Resolution[Resolution["resolved"] = 1] = "resolved";
    Resolution[Resolution["failed"] = 2] = "failed";
})(Resolution || (Resolution = {}));
function resolveLinks(aLinks, bLinks, entities) {
    if (aLinks.length !== bLinks.length)
        return Resolution.failed;
    var _loop_1 = function(a, v) {
        var resolved = entities[v];
        if (resolved === true) {
            return { value: Resolution.unknown };
        }
        else if (resolved === undefined) {
            throw new Error("Found a link for a non entity. " + [a, v]);
        }
        if (bLinks.some(function (_a) {
            var a2 = _a[0], v2 = _a[1];
            return a2 === a && v2 === resolved;
        }).length === 0) {
            return { value: Resolution.failed };
        }
    };
    for (var _i = 0, aLinks_1 = aLinks; _i < aLinks_1.length; _i++) {
        var _a = aLinks_1[_i], a = _a[0], v = _a[1];
        var state_1 = _loop_1(a, v);
        if (typeof state_1 === "object") return state_1.value;
    }
    return Resolution.resolved;
}
exports.resolveLinks = resolveLinks;
function resolveActualExpected(assert, actuals, expecteds, entities) {
    var ix = 0;
    var max = actuals.length * actuals.length;
    while (actuals[ix]) {
        var actual = actuals[ix];
        if (ix === max) {
            assert.true(false, "Cyclic test found");
            return;
        }
        ix++;
        var found = void 0;
        var expectedIx = 0;
        for (var _i = 0, expecteds_1 = expecteds; _i < expecteds_1.length; _i++) {
            var expected = expecteds_1[_i];
            var listEqual = void 0, linkEqual = void 0;
            if (isSetEqual(expected.list, actual.list, assert)) {
                listEqual = true;
            }
            else {
                found = false;
            }
            if (actual.links || expected.links) {
                var res = resolveLinks(actual.links, expected.links, entities);
                if (res === Resolution.failed) {
                    linkEqual = false;
                }
                else if (res === Resolution.resolved) {
                    linkEqual = true;
                }
                else {
                    linkEqual = false;
                    actuals.push(actual);
                    break;
                }
            }
            else {
                linkEqual = true;
            }
            if (listEqual && linkEqual) {
                expecteds.splice(expectedIx, 1);
                entities[actual.e] = expected.e;
                found = true;
                break;
            }
            expectedIx++;
        }
        if (found === false) {
            assert.true(false, "No matching add found for actual record: " + JSON.stringify(actual.list));
        }
    }
}
exports.resolveActualExpected = resolveActualExpected;
function verify(assert, adds, removes, data) {
    assert.equal(data.insert.length, adds.length, "Correct number of inserts");
    assert.equal(data.remove.length, removes.length, "Correct number of removes");
    // get all the entities
    var entities = collectEntities(adds);
    entities = collectEntities(data.insert, entities);
    entities = collectEntities(removes, entities);
    entities = collectEntities(data.remove, entities);
    //
    var expectedAdd = eavsToComparables(adds, entities);
    var expectedRemove = eavsToComparables(removes, entities);
    var actualRemove = eavsToComparables(data.remove, entities);
    var actualAdd = eavsToComparables(data.insert, entities);
    resolveActualExpected(assert, actualAdd, expectedAdd, entities);
    resolveActualExpected(assert, actualRemove, expectedRemove, entities);
}
exports.verify = verify;
function evaluate(assert, expected, code, session) {
    if (session === void 0) { session = new runtime_1.Database(); }
    var parsed = parser.parseDoc(dedent(code), "0");
    var _a = builder.buildDoc(parsed.results), blocks = _a.blocks, errors = _a.errors;
    if (expected.errors) {
        assert.true(parsed.errors.length > 0 || errors.length > 0, "This test is supposed to produce errors");
    }
    session.blocks = session.blocks.concat(blocks);
    var evaluation = new runtime_1.Evaluation();
    evaluation.registerDatabase("session", session);
    var changes = evaluation.fixpoint();
    verify(assert, expected.insert, expected.remove, changes.result());
    var next = { execute: function (expected, actions) {
            var changes = evaluation.executeActions(actions);
            verify(assert, expected.insert, expected.remove, changes.result());
            return next;
        }, session: session };
    return next;
}
exports.evaluate = evaluate;
;
function testSingleExpressionByList(list) {
    list.forEach(function (list_item, index) {
        test("Is " + list_item.expression + " returning " + list_item.expectedValue + "?", function (assert) {
            var expected = {
                insert: [
                    ["a", "floatTest", list_item.expectedValue],
                ],
                remove: [],
            };
            evaluate(assert, expected, "\n        ~~~\n        search\n          x = " + list_item.expression + "\n\n        bind\n          [floatTest: x]\n        ~~~\n      ");
            assert.end();
        });
    });
}
exports.testSingleExpressionByList = testSingleExpressionByList;
//# sourceMappingURL=shared_functions.js.map